<?php

/**
* Gateway class
**/
class WC_Gateway_Paymill extends WC_Payment_Gateway
{

	protected $live_api = null;
	protected $test_api = null;
	protected $test_private_key = null;
	protected $test_public_key = null;
	protected $live_private_key = null;
	protected $live_public_key = null;
	protected $logos = array();

	/**
	 * Constructor, set variables etc. and add hooks/filters
	 */
	function __construct() {
		$this->id                   = 'paymill';
		$this->method_title         = __( 'Paymill', 'wc_paymill' );
		$this->has_fields           = true;
		$this->live_api             = 'https://api.paymill.de/v2/';
		$this->test_api             = 'https://api.paymill.de/v2/';
		$this->supports             = array(
			'products',
			'refunds',
		);
		$this->view_transaction_url = 'https://app.paymill.com/transactions/%s';

		$this->available_currencies  = array(
			'AD', 'AT', 'BE', 'BG', 'CH', 'CY', 'CZ', 'DK',
			'DE', 'EE', 'ES', 'FO', 'FI', 'FR', 'GB', 'GR',
			'HR', 'HU', 'IE', 'IL', 'IM', 'IS', 'IT', 'LI',
			'LT', 'LU', 'LV', 'MC', 'MT', 'NL', 'NO', 'PL',
			'PT', 'RO', 'SE', 'SI', 'SK', 'TR', 'VA',
		);

		// Load the form fields
		$this->init_form_fields();

		// Load the settings.
		$this->init_settings();

		// Don't let the gateway be taken out of test mode unless
		// live API keys have been provided
		if ( empty( $this->settings['live_public_key'] ) ||
			 empty( $this->settings['live_private_key'] ) ) {
			if ( 'yes' == $this->settings['testmode'] ) {
				$this->form_fields['testmode']['description'] .= __( "<br><strong>Note</strong>: You cannot go live until you've provided live public and private API keys", 'wc_paymill' );
			}
		}
		// If people are upgrading from a release without checkout mode choice, preserve the
		// old style form, not the frame-based form.
		if ( empty( $this->settings['checkout_mode'] ) &&
			 ( ! empty( $this->settings['live_private_key'] ) ||
			   ! empty( $this->settings['test_private_key'] ) )
			   ) {
			// The user has a keys, but no checkout mode. Assume they want to keep the old one.
			$this->settings['checkout_mode'] = 'embedded';
		}

		// Get setting values
		$this->title            = $this->settings['title'];
		$this->description      = $this->settings['description'];
		$this->debug            = ! empty( $this->settings['debug'] ) ? $this->settings['debug'] : '';
		$this->enabled          = $this->settings['enabled'];
		$this->testmode         = $this->settings['testmode'];
		$this->test_private_key = $this->settings['test_private_key'];
		$this->test_public_key  = $this->settings['test_public_key'];
		$this->live_private_key = $this->settings['live_private_key'];
		$this->live_public_key  = $this->settings['live_public_key'];
		$this->checkout_mode    = $this->settings['checkout_mode'];
		$this->logos            = $this->parse_logo_settings();

		// Set up localisation
		add_action( 'init', array( $this, 'init' ) );

		// Check for SSL and warn if not present
		add_action( 'admin_notices', array( $this, 'ssl_check' ) );
		add_action( 'admin_notices', array( $this, 'test_mode_check' ) );
		add_action( 'admin_notices', array( $this, 'base_country_check' ) );

		// Save options from admin forms
		add_action( 'woocommerce_update_options_payment_gateways', array( $this, 'process_admin_options' ) ); // WC < 2.0
		add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, array( $this, 'process_admin_options' ) ); // WC 2.0+

		// Enqueue javascript and CSS
		add_action( 'wp_enqueue_scripts', array( $this, 'enqueue_scripts' ) );
		add_action( 'wp_enqueue_scripts', array( $this, 'enqueue_styles' ) );

		add_action( 'woocommerce_after_my_account', array( $this, 'payment_card_admin' ) );

		add_filter( 'woocommerce_gateway_icon', array( $this, 'show_icons' ), 10, 2 );
	}

	/**
	 * Set up translation stuff
	 */
	public function init() {
		$locale = apply_filters( 'plugin_locale', get_locale(), 'wc_paymill' );
		load_textdomain( 'wc_paymill', WP_LANG_DIR . '/wc_paymill/wc_paymill-' . $locale . '.mo' );
		load_plugin_textdomain( 'wc_paymill', false, basename( dirname( __FILE__ ) ) . '/languages/' );
	}

	/**
	 * Log information using the WC_Logger class.
	 *
	 * Will do nothing unless debug is enabled.
	 */
	private function log( $msg ) {
		static $logger = false;
		// Bail if debug isn't on.
		if ( 'yes' != $this->debug ) {
			return;
		}
		// Create a logger instance if we don't already have one.
		if ( false === $logger ) {
			$logger = new WC_Logger();
		}
		$logger->add( 'paymill', $msg );
	}

	/**
	 * Populate an easy to parse array based on the logo settings
	 *
	 * @return array An array of which card logos are required
	 */
	private function parse_logo_settings() {
		$logos = array();
		foreach ( $this->settings as $setting => $value ) {
			if ( stripos( $setting, 'card-' ) === 0 && 'yes' == $value ) {
				$card_type = substr( $setting, 5 );
				$logos[ $card_type ] = $value;
			}
		}
		return $logos;
	}

	/**
	 * From:
	 * https://github.com/skyverge/wc-plugin-compatibility/
	 */
	private function get_wc_version() {
		if ( defined( 'WC_VERSION' )          && WC_VERSION )          return WC_VERSION;
		if ( defined( 'WOOCOMMERCE_VERSION' ) && WOOCOMMERCE_VERSION ) return WOOCOMMERCE_VERSION;
		return null;
	}

	/**
	 * From:
	 * https://github.com/skyverge/wc-plugin-compatibility/
	 */
	private function is_wc_2_1() {
		return version_compare( $this->get_wc_version(), '2.1-beta', '>' );
	}

	/**
	 * Make an SSL version of a URL if required.
	 * From: https://github.com/skyverge/wc-plugin-compatibility/
	 */
	private function ssl_url( $url ) {
		if ( $this->is_wc_2_1() ) {
			return WC_HTTPS::force_https_url( $url );
		} else {
			global $woocommerce;
			return $woocommerce->force_ssl( $url );
		}
	}

	/**
	 * Output the logos selected
	 * @param  string $icon    The default WC-generated icon
	 * @param  string $gateway The gateway the icons are for
	 * @return string          The HTML for the selected iconsm or empty string if none
	 */
	public function show_icons( $icon, $gateway ) {
		if ( 'paymill' != $gateway ) {
			return $icon;
		}
		$output = '';
		foreach ( array_keys( $this->logos ) as $provider ) {
			$url = plugins_url( basename( plugin_dir_path(__FILE__) ), basename( __FILE__ ) ) . "/images/card-$provider.png";
			$classes = 'wpec_paymill_logo wpec_paymill_logo_' . esc_attr($provider);
			$output .= '<img class="'.$classes.'" src="' . $this->ssl_url( $url ) . '" alt="' . $this->title . '" />';
		}
		return $output;
	}

	/**
	 * Check if the store's base country is supported by Paymill
	 *
	 * @return boolean True if the base country is supported
	 */
	private function is_base_country_supported() {
		global $woocommerce;
		return in_array( $woocommerce->countries->get_base_country(), $this->available_currencies );
	}

	/**
	 * Check if we have API keys for the chosen mode (Test or live)
	 *
	 * @return  boolean Return true if we have keys for the chosen mode
	 */
	private function have_api_keys() {
		if ( $this->settings['testmode'] == 'yes' ) {
			if ( ! empty ( $this->settings['test_public_key'] ) &&
				 ! empty ( $this->settings['test_private_key'] ) ) {
				return true;
			} else {
				return false;
			}
		} else {
			if ( ! empty ( $this->settings['live_public_key'] ) &&
				 ! empty ( $this->settings['live_private_key'] ) ) {
				return true;
			} else {
				return false;
			}
		}
	}

	/**
	 * Get the API endpoint in use
	 * @return string API endpoint URL
	 */
	private function get_api() {
		if ( $this->testmode == 'yes' ) {
			$api = $this->test_api;
		} else {
			$api = $this->live_api;
		}
		return $api;
	}

	/**
	 * Retrieve the public key for either live or test depending on settings.
	 *
	 * @return string  The public API key.
	 */
	private function get_public_key() {
		if ( $this->testmode == 'yes' ) {
			return $this->test_public_key;
		} else {
			return $this->live_public_key;
		}
	}

	/**
	 * Get the private API key to use
	 * @return string The private API key to use
	 */
	private function get_private_key() {
		if ( $this->testmode == 'yes' ) {
			$private_key = $this->settings['test_private_key'];
			$api = $this->test_api;
		} else {
			$private_key = $this->settings['live_private_key'];
			$api = $this->live_api;
		}
		return $private_key;
	}

	/**
	 * Retrieves the paymill client ID from the user record if there is one, or create one.
	 * Returns null if there is no logged in user, or the client ID can't be created
	 * @param bool $create_client Whether to create a Paymill client if one isn't already registered
	 * @return object A paymill user_meta object, or null
	 */
	protected function get_paymill_user_meta( $create_client = true ) {
		$current_user = wp_get_current_user();
		// We need a user, and an email address to create a client ID
		if ( $current_user->ID == 0 || empty ( $current_user->data->user_email ) )
			return null;

		$user_meta = get_user_meta( $current_user->ID, '_wc_paymill', true );

		// Grab existing ID if there is one
		if ( ! empty ( $user_meta ) )
			return $user_meta;

		// If we get here then the user doesn't have Paymill client info stored against their
		// user account

		// Don't create an account if we were told not to
		if ( ! $create_client )
			return null;

		// Create a client ID and store it in user meta
		$private_key = $this->get_private_key();
		$api = $this->get_api();

		try {
			$clientObject = new Services_Paymill_Clients( $private_key, $api );
			$client = $clientObject->create( array( 'email' => $current_user->data->user_email ) );
			if ( ! empty ( $client['id'] ) ) {
				$user_meta = new stdClass();
				$user_meta->client_id = $client['id'];
				add_user_meta( $current_user->ID, '_wc_paymill', $user_meta, true );
				$this->log( 'Client created successfully: ' . $client['id'] );
				return $user_meta;
			}
		} catch ( Exception $e ) {
			$this->log( 'Error creating client: ' . $e->getMessage() );
			return null;
		}
	}

	/**
	 * Remove data from the _wc_paymill user meta store
	 * @param  string $type The type of data being stored - this is used as a class member
	 * @param  mixed $key The key of the data to remove
	 * @return void
	 */
	protected function remove_paymill_user_meta( $type, $key ) {
		$current_user = wp_get_current_user();
		if ( $current_user->ID == 0 ) {
			return null;
		}
		$user_meta = get_user_meta( $current_user->ID, '_wc_paymill', true );
		if ( ! isset ( $user_meta->{$type}[$key] ) ) {
			return;
		}
		unset ( $user_meta->{$type}[$key] );
		update_user_meta( $current_user->ID, '_wc_paymill', $user_meta );
	}

	/**
	 * Store data in the _wc_paymill user meta store
	 * @param  string $type The type of data being stored - this is used as a class member
	 * @param  mixed $data The data to store
	 * @return void
	 */
	protected function store_paymill_user_meta( $type, $data ) {
		$current_user = wp_get_current_user();
		if ( $current_user->ID == 0 ) {
			return null;
		}
		$user_meta = get_user_meta( $current_user->ID, '_wc_paymill', true );

		if ( ! isset( $user_meta->$type ) ) {
			$user_meta->$type = $data;
		} else {
			$user_meta->$type = array_merge( $user_meta->$type, $data );
		}
		update_user_meta( $current_user->ID, '_wc_paymill', $user_meta );
	}

	/**
	 * Generate a user-visible error message based on a transaction response
	 * @param  array $transaction A transaction reponse
	 * @return string             The user visible error message
	 */
	private function parse_transaction_error( $transaction ) {
		if ( ! isset( $transaction['response_code'] ) ) {
			$msg =  __( 'There has been a problem processing your payment, please try again.', 'wc_paymill' );
		} else {
			switch ( $transaction['response_code'] ) {
				case '40000':
				case '40101':
				case '40102':
				case '40105':
				case '40200':
				case '40300':
					$msg = __( 'There is a problem with your credit card data, please check your details and try again.', 'wc_paymill' );
					break;
				case '40100':
					$msg = __( 'Please double-check your card information. If you receive this error with correct card information, it generally means that you need to contact your issuing bank to have your charge pre-authorised.', 'wc_paymill' );
					break;
				case '40103':
				case '50101':
				case '50102':
					$msg = __( 'Insufficient funds. Please provide alternate payment details.', 'wc_paymill' );
					break;
				case '40104':
				case '50105':
				case '50200':
					$msg = __( 'Please provide alternate payment details.', 'wc_paymill' );
					break;
				case '50400':
					$msg = __( 'Please try another form of payment, or contact your bank if this error occurs more than once.', 'wc_paymill' );
					break;
				case '50001':
				case '50103':
				case '50104':
				case '50201':
					$msg = __( 'There is a problem with your credit card data, please contact your issuing bank.', 'wc_paymill' );
					break;
				case '40301':
				case '40401':
				case '40402':
				case '40403':
					$msg = __( 'There has been a problem taking your payment, please contact the store owner to complete payment.', 'wc_paymill' );
					break;
				case '40400':
				case '50000':
				case '50100':
				case '50300':
				default:
					if (!empty($transaction['error'])) {
						$msg = $transaction['error'];
					} else {
						$msg = __( 'There was a problem processing your payment. Please try again.', 'wc_paymill' );
					}
					break;
			}
		}
		return apply_filters( 'wc_paymill_user_error_message', $msg, $transaction['response_code'] );
	}

	/**
	 * Process a payment using a Paymill token
	 * @param  array $args The arguments to pass to Paymill with the transaction details
	 * @return array       Paymill status response - or null
	 */
	protected function process_payment_by_token( $args ) {
		$private_key = $this->get_private_key();
		$api = $this->get_api();
		try {
			$transactionsObject = new Services_Paymill_Transactions( $private_key, $api );
			$transaction = $transactionsObject->create( $args );
			if ( isset( $transaction['error'] ) || ! isset( $transaction['id'] ) ) {
				$response = $transactionsObject->getResponse();
				if (!empty($response['body']['data'])) {
					$transaction = $response['body']['data'];
				} else {
					if ( !empty( $response['body']['error'] ) ) {
						throw new Exception( $response['body']['error'] );
					} else {
						throw new Exception(
							__( 'There was a problem processing your payment. Please try again.', 'wc_paymill' )
						);
					}
				}
				throw new Exception( $this->parse_transaction_error( $transaction ) );
			}
			$this->log( 'Transaction created successfully: ' . $transaction['id'] );
			return $transaction;
		} catch ( Exception $e ) {
			$this->log( 'Error creating transaction in process_payment_by_token(): ' . $e->getMessage() );
			wc_add_notice( $e->getMessage(), 'error' );
			return $transaction;
		}
	}

	/**
	 * Process a payment against a client account
	 * @param  array $args The arguments to pass to Paymill with the transaction details
	 * @param  object $paymill_user_meta A Paymill user meta object
	 * @return array or null       Paymill status response - or null
	 */
	protected function process_payment_for_client( $args, $paymill_user_meta ) {
		global $woocommerce;
		$private_key = $this->get_private_key();
		$api = $this->get_api();
		$transaction = array();
		try {
			// If this is a new card, then create a payment
			if ( empty( $args['payment'] ) ) {
				$paymentsObject = new Services_Paymill_Payments( $private_key, $api );
				$payment_args = array( 'client' => $paymill_user_meta->client_id, 'token' => $args['token'], 'source' => 'woothemes' );
				$payment = $paymentsObject->create( $payment_args );
				if ( empty( $payment['id'] ) ) {
					throw new Exception( __( 'There is a problem with your credit card data, please check your details and try again', 'wc_paymill' ) );
				}
				$this->log( 'Payment created successfully: ' . $payment['id'] );
				// Add the payment ID to the transaction request
				$args['payment'] = $payment['id'];

				// Store the payment in the user meta
				$this->store_paymill_user_meta( 'payments', array( $payment['id'] => $payment ) );
			}
			// Make the transaction with the payment and client IDs
			$transactionsObject = new Services_Paymill_Transactions( $private_key, $api );
			unset ( $args['token'] );
			$args['client'] = $paymill_user_meta->client_id;
			$transaction = $transactionsObject->create( $args );
			if ( isset( $transaction['error'] ) || ! isset( $transaction['id'] ) ) {
				$response = $transactionsObject->getResponse();
				$transaction = $response['body']['data'];
				throw new Exception( $this->parse_transaction_error( $transaction ) );
			}
			$this->log( 'Transaction created successfully: ' . $transaction['id'] );
			return $transaction;
		} catch( Exception $e ) {
			$this->log( 'Error creating transaction in process_payment_for_client(): ' . $e->getMessage() );
			wc_add_notice( $e->getMessage(), 'error' );
			return $transaction;
		}
	}

	/**
	* Check if SSL is enabled and notify the user
	**/
	function ssl_check() {
		if ( $this->enabled == 'no' ) {
			return;
		}
		// Show message if enabled and FORCE SSL is disabled and WordpressHTTPS plugin is not detected
		if ( get_option( 'woocommerce_force_ssl_checkout' ) == 'no' && !class_exists( 'WordPressHTTPS' ) ) {
			echo '<div class="error"><p>'.sprintf( __( 'Paymill payments are enabled, but the <a href="%s">Force secure option</a> is disabled; your checkout may not be secure! Please enable SSL and ensure your server has a valid SSL certificate. For further information, please see the question &quot;Do I need SSL encryption to use Paymill?&quot; <a href="%s">here</a>', 'wc_paymill' ), admin_url( 'admin.php?page=woocommerce' ), 'https://www.paymill.com/sl-si/support-3/worth-knowing/faqs/index.html').'</p></div>';
		}
	}

	/**
	 * Check if test mode is enabled, and display a warning in the admin area if it is
	 */
	function test_mode_check() {
		if ( $this->enabled == 'no' ) {
			return;
		}
		if ( $this->settings['testmode'] == 'yes' ) {
			echo '<div class="error"><p>'.__( 'The Paymill gateway is enabled, but is operating in <strong>test mode</strong>. No real funds will be transferred.', 'wc_paymill' ).'</p></div>';
		}
	}

	/**
	 * Check that the currency and country are supported by Paymill and notify the user if not
	 */
	function base_country_check() {
		global $woocommerce;
		if ( $this->enabled == 'no' ) {
			return;
		}
		if ( ! $this->is_base_country_supported() ) {
			echo '<div class="error"><p>'.sprintf( __( 'Paymill payments are enabled, but Paymill do not currently support %s as a base country. For further information, please see the <a href="%s">list of supported countries</a>', 'wc_paymill' ), $woocommerce->countries->countries[$woocommerce->countries->get_base_country()], 'https://support.paymill.com/hc/en-us/articles/200262037-Which-currencies-countries-does-PAYMILL-accept-and-pay-out-').'</p></div>';
			return;
		}
	}

	/**
	 * Enqueue needed javascript
	 */
	function enqueue_scripts() {
		$script_suffix = defined( 'SCRIPT_DEBUG' ) && SCRIPT_DEBUG ? '' : '.min';
		if ( is_checkout() ) {
			wp_enqueue_script( 'wc-paymill-bridge', 'https://bridge.paymill.com', array() );
			if ( $this->checkout_mode == 'frame' ) {
				$script_path = plugins_url(
					basename( plugin_dir_path( __FILE__ ) ),
					basename( __FILE__ )
				) . '/js/paymill-checkout-frame' . $script_suffix . '.js';
			} else {
				$script_path = plugins_url(
					basename( plugin_dir_path( __FILE__ ) ),
					basename( __FILE__ )
				) . '/js/paymill-checkout' . $script_suffix . '.js';
			}
			wp_enqueue_script( 'wc-paymill', $script_path, array( 'jquery', 'woocommerce' ) );
		}
		$locale = get_locale();
		$locale = strtolower(substr($locale, 0, 2));
		if (!in_array($locale, array('en', 'de', 'fr', 'it', 'es', 'pt', 'da', 'hu'))) {
			$locale = 'en';
		}
		$translatable_script_data = array(
			'threeds_cancel_label' => __( 'Cancel', 'woocommerce_gpf' ),
			'locale' => $locale,
		);
		wp_localize_script( 'wc-paymill', 'wc_paymill', $translatable_script_data );
	}

	/**
	 * Enqueue needed CSS
	 */
	function enqueue_styles() {
		$style_path = plugins_url( basename( plugin_dir_path( __FILE__ ) ), basename( __FILE__ ) ) . '/css/paymill-checkout.css';
		if ( is_checkout() ) {
			wp_enqueue_style( 'wc-paymill', $style_path, array() );
		}
	}

	/**
	 * Initialise Gateway Settings Form Fields
	 */
	function init_form_fields() {
		$this->form_fields = array(
			'enabled' => array(
				'title' => __( 'Enable/Disable', 'wc_paymill' ),
				'label' => __( 'Enable Paymill', 'wc_paymill' ),
				'type' => 'checkbox',
				'description' => '',
				'default' => 'no',
			),
			'debug' => array(
				'title' => __( 'Enable debug mode', 'wc_paymill' ),
				'label' => __( 'Enable only if you are diagnosing problems.', 'wc_paymill' ),
				'type' => 'checkbox',
				'description' => sprintf( __( 'Log Paymill interactions inside <code>%s</code>', 'wc_paymill' ), wc_get_log_file_path( 'paymill' ) ),
				'default' => 'no',
			),
			'title' => array(
				'title' => __( 'Title', 'wc_paymill' ),
				'type' => 'text',
				'description' => __( 'This controls the description which the user sees during checkout.', 'wc_paymill' ),
				'default' => __( 'Credit card (Paymill)', 'wc_paymill' )
			),
			'description' => array(
				'title' => __( 'Description', 'wc_paymill' ),
				'type' => 'textarea',
				'description' => __( 'This controls the description which the user sees during checkout.', 'wc_paymill' ),
				'default' => 'Pay with your credit card via Paymill.',
			),
			'checkout_mode' => array(
				'title' => __( 'Checkout mode', 'wc_paymill' ),
				'description' => __( 'Choose which checkout form to use. This affects the level of PCI compliance you will need to go through. "Frame" has the lower PCI compliance requirements.', 'wc_paymill' ),
				'type' => 'select',
				'default' => '',
				'options' => array(
					'frame' => 'Frame (Lower PCI compliance requirements. Recommended)',
					'embedded' => 'On-site form.'
				),
			),
			'testmode' => array(
				'title' => __( 'Test Mode', 'wc_paymill' ),
				'label' => __( 'Enable Paymill Test Mode', 'wc_paymill' ),
				'type' => 'checkbox',
				'description' => __( 'Choose whether the payment gateway should be in development / test mode.', 'wc_paymill' ),
				'default' => 'yes',
			),
			'card-amex' => array(
				'title' => __( 'Card Logos', 'wc_paymill' ),
				'checkboxgroup' => 'start',
				'label' => __( 'American Express', 'wc_paymill' ),
				'type' => 'checkbox',
				'description' => '',
				'default' => 'no',
			),
			'card-diners' => array(
				'title' => '',
				'label' => __( 'Diners Club', 'wc_paymill' ),
				'type' => 'checkbox',
				'description' => '',
				'default' => 'no',
			),
			'card-jcb' => array(
				'title' => '',
				'label' => __( 'JCB', 'wc_paymill' ),
				'type' => 'checkbox',
				'description' => '',
				'default' => 'no',
			),
			'card-maestro' => array(
				'title' => '',
				'label' => __( 'Maestro', 'wc_paymill' ),
				'type' => 'checkbox',
				'description' => '',
				'default' => 'no',
			),
			'card-mastercard' => array(
				'title' => '',
				'label' => __( 'Mastercard', 'wc_paymill' ),
				'type' => 'checkbox',
				'description' => '',
				'default' => 'yes',
			),
			'card-visa' => array(
				'title' => '',
				'label' => __( 'Visa', 'wc_paymill' ),
				'checkboxgroup' => 'end',
				'type' => 'checkbox',
				'description' => '',
				'default' => 'yes',
			),
			'test_details' => array(
				'title' => __( 'Test API details', 'wc_paymill' ),
				'type' => 'title',
				'description' => ''
			),
			'test_private_key' => array(
				'title' => __( 'Private Key', 'wc_paymill' ),
				'type' => 'text',
				'description' => __( 'Your test private key.', 'wc_paymill' ),
				'default' => '',
			),
			'test_public_key' => array(
				'title' => __( 'Public Key', 'wc_paymill' ),
				'type' => 'text',
				'description' => __( 'Your test public key.', 'wc_paymill' ),
				'default' => '',
			),
			'live_details' => array(
				'title' => __( 'Live API details', 'wc_paymill' ),
				'type' => 'title',
				'description' => ''
			),
			'live_private_key' => array(
				'title' => __( 'Private Key', 'wc_paymill' ),
				'type' => 'text',
				'description' => __( 'Your live private key.', 'wc_paymill' ),
				'default' => '',
			),
			'live_public_key' => array(
				'title' => __( 'Public Key', 'wc_paymill' ),
				'type' => 'text',
				'description' => __( 'Your live public key.', 'wc_paymill' ),
				'default' => '',
			),
		);
	}

	/**
	 * Admin Panel
	 */
	function admin_options() {
		?>
		<h3><?php _e( 'Paymill', 'wc_paymill' ); ?></h3>
		<p><?php _e( 'Paymill works by adding credit card fields on the checkout and then sending the details to Paymill for verification and processing.', 'wc_paymill' ); ?></p>
		<table class="form-table">
			<?php $this->generate_settings_html(); ?>
		</table><!--/.form-table-->
		<?php
	}

	/**
	 * Allow the user to remove existing card details
	 * @return void
	 */
	function payment_card_admin() {
		// Remove cards if requested.
		if ( ! empty ( $_POST['wc_paymill_delete_card'] ) && wp_verify_nonce( $_POST['_wpnonce'], 'wc_paymill_delete_card' ) ) {
			$this->remove_paymill_user_meta( 'payments', $_POST['wc_paymill_delete_card'] );
		}
		$paymill_user_meta = $this->get_paymill_user_meta( false );
		if ( empty ( $paymill_user_meta->payments ) ) {
			return;
		}
		?>
		<h2 id="saved-cards" style="margin-top:40px;"><?php _e( 'Saved cards', 'wc_paymill' ); ?></h2>
		<table class="shop_table">
			<thead>
				<tr>
					<th><?php _e( 'Card type', 'wc_paymill' ); ?></th>
					<th><?php _e( 'Ending in', 'wc_paymill' ); ?></th>
					<th><?php _e( 'Expires', 'wc_paymill' ); ?></th>
				</tr>
			</thead>
			<tbody>
				<?php foreach ( $paymill_user_meta->payments as $credit_card ) : ?>
				<tr>
					<td><?php esc_html_e( ucfirst( $credit_card['card_type'] ) ); ?></td>
					<td><?php esc_html_e( $credit_card['last4'] ); ?></td>
					<td><?php echo esc_html( $credit_card['expire_month'] ) . '/' . esc_html( $credit_card['expire_year'] ); ?></td>
					<td>
						<form action="#saved-cards" method="post">
							<?php wp_nonce_field( 'wc_paymill_delete_card' ); ?>
							<input type="hidden" name="wc_paymill_delete_card" value="<?php esc_attr_e( $credit_card['id'] ); ?>">
							<input type="submit" name="submit" value="<?php _e( 'Remove', 'wc_paymill' ); ?>">
						</form>
					</td>
				</tr>
				<?php endforeach; ?>
			</tbody>
		</table>
		<?php
	}

	/**
	 * Check if this gateway is enabled and available in the user's country
	 */
	function is_available() {
		if ( $this->enabled == 'yes' &&
			 $this->is_base_country_supported() &&
			 $this->have_api_keys() ) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Output the payment form on the checkout page
	 */
	function payment_fields() {
		global $woocommerce;
		$payment_amount = $woocommerce->cart->total;
		if ( $this->testmode == 'yes' ) {
			echo '<p class="woocommerce_error">'.__( 'TEST MODE/SANDBOX ENABLED', 'wc_paymill' ).'</p>';
		}

		if ( $this->description ) {
			echo '<legend>'.esc_html( $this->description ).'</legend>';
		}
		$public_key = $this->get_public_key();
		?>
		<script type="text/javascript">
			var PAYMILL_PUBLIC_KEY = '<?php esc_attr_e( $public_key ); ?>';
			var PAYMILL_CURRENCY   = '<?php esc_attr_e( get_woocommerce_currency() ); ?>';
		</script>

		<fieldset id="paymill_ccform">
			<?php
		$paymill_user_meta = $this->get_paymill_user_meta();

		if ( ! empty( $paymill_user_meta->payments ) ) {
			?>
			<div>
				<p>
					<select name="paymill_payment_id" id="paymill_payment_id"><?php
			foreach ( $paymill_user_meta->payments as $payment ) {
				echo '<option value="'.esc_attr( $payment['id'] ).'">';
				echo 'xxxx-xxxx-xxxx-'.esc_html( $payment['last4'] );
				echo ' ('.esc_html( $payment['card_type'] ).') ';
				echo ' - '.esc_html( $payment['card_holder'] );
				echo '</option>';
			}
			?>
						<option value=""><?php _e( 'Use a new card', 'wc_paymill' ); ?></option>
					</select>
				</p>
			</div><?php
		} else {
			?>
			<input type="hidden" name="paymill_payment_id" id="paymill_payment_id" value="">
			<?php
		}
		?>
			<div id="paymill_new_card_form" class="paymill-<?php echo esc_attr( $this->checkout_mode ); ?>">
		<?php
		if ($this->checkout_mode != 'frame') {
			$this->embedded_payment_fields($payment_amount);
		}
		?>
			</div>
			<div class="clear"></div>
		</fieldset>
		<div style="display: none;" id="paymill_messages" class="woocommerce_error"><?php _e( 'There are errors with your card details, please double check them and try again.', 'wc_paymill' ); ?></div>
		<?php
			// Output the order total so our JS can get it
			// Required for 3d-secure payments.
		?>
			<input type="hidden" id="paymill_order_total" value="<?php echo esc_attr( $payment_amount * 100 ); ?>">
		<?php
	}

	/**
	 * Render the payment for for on-site based collection.
	 */
	private function embedded_payment_fields($payment_amount) {
		global $wp_locale;
			?>
				<p class="form-row form-row-first">
					<label for="paymill_card_number"><?php _e( 'Credit Card number', 'wc_paymill' ) ?> <span class="required">*</span></label>
					<input type="text" class="input-text paymill_checkout_field" id="paymill_card_number" />
				</p>
				<p class="form-row form-row-first">
					<label for="paymill_cardholder"><?php _e( 'Card holder name', 'wc_paymill' ) ?> <span class="required">*</span></label>
					<input type="text" class="input-text paymill_checkout_field" id="paymill_cardholder" />
				</p>
				<p class="form-row form-row-first">
					<label for="paymill_card_expiry_month"><?php _e( 'Expiration date', 'wc_paymill' ); ?> <span class="required">*</span></label>
					<select id="paymill_card_expiry_month" class="woocommerce-select woocommerce-cc-month paymill_checkout_field" >
						<option value=""><?php _e( 'Month', 'wc_paymill' ); ?></option>
						<?php
							$months = array();
							for ($i = 1; $i <= 12; $i++) :
								$timestamp = mktime( 0, 0, 0, $i, 1 );
								$months[date( 'm', $timestamp )] = $wp_locale->get_month( date( 'm', $timestamp ) );
							endfor;
							foreach ( $months as $num => $name ) {
								printf( '<option value="%s">%s</option>', $num, $name );
							}
						?>
					</select>
					<select id="paymill_card_expiry_year" class="woocommerce-select woocommerce-cc-year paymill_checkout_field" >
						<option value=""><?php _e( 'Year', 'wc_paymill' ); ?></option>
						<?php
							for ( $i = date( 'Y' ); $i <= date( 'Y' ) + 15; $i++ ) {
								printf( '<option value="%u">%u</option>', $i, $i );
							}
						?>
					</select>
				</p>
				<p class="form-row form-row-first">
					<label for="paymill_cvv"><?php _e( 'Card security code', 'wc_paymill' ); ?> <span class="required">*</span></label>
					<input type="text" class="input-text paymill_checkout_field" id="paymill_cvv" maxlength="4" style="width:4em;" />
				</p>
		<?php
	}

	/**
	 * Validate that a valid payment token has been received.
	 * NOTE: Most validation happens on the browser side with the paymill API, see paymill-checkout.js
	 */
	function validate_fields() {
		global $woocommerce;
		if ( ! empty ( $_POST['paymill_token'] ) || ! empty ( $_POST['paymill_payment_id'] ) ) {
			return true;
		} else {
			$this->log( 'validate_fields() failed as no token, or payment ID were provided' );
			wc_add_notice( __( 'Payment details are invalid', 'wc_paymill' ), 'error' );
			return false;
		}
	}

	/**
	 * Process the payment
	 */
	function process_payment( $order_id ) {
		global $woocommerce;
		$order = new WC_Order( $order_id );
		$paymill_user_meta = $this->get_paymill_user_meta();
		$arguments = array(
			'currency'    => get_woocommerce_currency(),
			'token'       => !empty( $_POST['paymill_token'] ) ? $_POST['paymill_token'] : null,
			'description' => sprintf( __( 'Order number %s' ), $order->get_order_number() ),
			'source'      => 'woothemes',
		);
		// Add the order total.
		// NOTE - amount is in "cents" or equivalent
		if ( is_callable( array( $order, 'get_total' ) ) ) {
			// WooCommerce 2.7+
			$arguments['amount'] = $order->get_total() * 100;
		} else {
			$arguments['amount'] = $order->order_total * 100;
		}
		if ( ! empty( $paymill_user_meta ) ) {
			$arguments['payment'] = isset( $_POST['paymill_payment_id'] ) ? $_POST['paymill_payment_id'] : '';
			$transaction = $this->process_payment_for_client( $arguments, $paymill_user_meta );
		} else {
			// Otherwise process the payment using the token
			$transaction = $this->process_payment_by_token( $arguments );
		}
		$transaction_status = isset( $transaction['status'] ) ? $transaction['status'] : 'card create';
		// If the payment API call fails we'll have an empty error as $transaction. Populate values we need
		if ( empty ( $transaction ) ) {
			$transaction['id'] = '';
			$transaction['payment']['id'] = '';
			$transaction['response_code'] = 'CARD CREATE FAILURE';
			$transaction['status'] = 'CARD CREATE FAILURE';
		}
		// Check the status
		switch ( $transaction_status ) {
			// Successful payment
			case 'closed':
				// Add order note
				$order->add_order_note(
					sprintf(
						__( 'Paymill payment completed (Transaction ID: %s, Payment ID: %s)', 'wc_paymill' ),
						$transaction['id'],
						$transaction['payment']['id']
					)
				);
				$this->log(
					sprintf(
						__( 'Paymill payment completed (Transaction ID: %s, Payment ID: %s)', 'wc_paymill' ),
						$transaction['id'],
						$transaction['payment']['id']
					)
				);

				// Payment complete
				$order->payment_complete( $transaction['id'] );

				// Remove cart
				$woocommerce->cart->empty_cart();

				if ( method_exists( $order, 'get_checkout_order_received_url' ) ) {
					$redirect = $order->get_checkout_order_received_url();
				} else {
					if ( is_callable( array( $order, 'get_id' ) ) &&
					     is_callable( array( $order, 'get_order_key' ) ) ) {
						$redirect = add_query_arg( 'key', $order->get_order_key(), add_query_arg( 'order', $order->get_id(), get_permalink( get_option( 'woocommerce_thanks_page_id' ) ) ) );
					} else {
						$redirect = add_query_arg( 'key', $order->order_key, add_query_arg( 'order', $order->id, get_permalink( get_option( 'woocommerce_thanks_page_id' ) ) ) );
					}
				}

				// Return thank you page redirect
				return array(
					'result'    => 'success',
					'redirect'  => $redirect,
				);
				break;

			default:
				$this->log(
					sprintf(
						__( 'Paymill payment failed (Transaction ID: %s, Payment ID: %s). Payment status %s. (%s)', 'wc_paymill' ),
						$transaction['id'],
						$transaction['payment']['id'],
						$transaction['status'],
						$transaction['response_code']
					)
				);
				// Payment status uncertain
				$order->update_status( 'failed', sprintf( __( 'Paymill payment failed (Transaction ID: %s, Payment ID: %s). Payment status %s. (%s)', 'wc_paymill' ), $transaction['id'], $transaction['payment']['id'], $transaction['status'], $transaction['response_code'] ) );
				return;
				break;
		}
		return;
	}

	/**
	 * Allow refunds to be processed from WooCommerce.
	 * @param  int            $order_id  WooCommerce order ID.
	 * @param  int            $amount    Amount to be refunded.
	 * @param  string         $reason    Refund description
	 * @return bool|WP_Error             True on success, false on failure.
	 */
	public function process_refund( $order_id, $amount = null, $reason = null ) {
		if ( !$amount ) {
			return false;
		}
		if ( is_null( $reason ) ) {
			$reason = 'Refund relating to order #' . $order_id;
		}
		$order = new WC_Order( $order_id );
		$transaction_id = $order->get_transaction_id( $order );
		$private_key = $this->get_private_key();
		$api = $this->get_api();
		$refund = new Services_Paymill_Refunds( $private_key, $api );
		$params = array(
			'amount'      => $amount * 100,
			'description' => $reason,
			'source'      => 'woothemes'
		);
		$itemdata = array(
			'transactionId' => $transaction_id,
			'params'        => $params,
		);
		try {
			$results = $refund->create( $itemdata );
			if ( ! isset($results['data']['response_code']) || $results['data']['response_code'] != '20000' ) {
				return false;
			}
			$refund_id = $results['data']['id'];
			$order->add_order_note(
				sprintf(
					__( 'Refund of %.2f processed (Transaction ID: %s)', 'wc_paymill' ),
					$amount,
					$refund_id
				)
			);
			return true;
		} catch ( Exception $e ) {
			return false;
		}
	}
}
